@@ -7,13 +7,15 @@ from importlib.abc import ExecutionLoader
 import importlib.machinery
 import logging
 import os
+from pathlib import Path, PurePath
 import sys
 from types import CodeType
-from typing import Dict, List, Optional, Tuple, Union
+from typing import Any, Dict, List, Optional, Set, Tuple, Union
 import opcode
 
-from cx_Freeze.common import code_object_replace
-from cx_Freeze.module import Module
+from .common import code_object_replace, process_path_specs
+from .common import IncludesList, InternalIncludesList
+from .module import ConstantsModule, Module
 
 
 BUILD_LIST = opcode.opmap["BUILD_LIST"]
@@ -33,19 +35,23 @@ __all__ = ["Module", "ModuleFinder"]
 
 
 class ModuleFinder:
+    """ModuleFinder base class."""
+
     def __init__(
         self,
-        include_files: Optional[List[str]] = None,
+        include_files: Optional[IncludesList] = None,
         excludes: Optional[List[str]] = None,
-        path: Optional[str] = None,
-        replace_paths: Optional[List[str]] = None,
+        path: Optional[List[str]] = None,
+        replace_paths: Optional[List[Tuple[str, str]]] = None,
         zip_include_all_packages: bool = False,
         zip_exclude_packages: Optional[List[str]] = None,
         zip_include_packages: Optional[List[str]] = None,
-        constants_module=None,
-        zip_includes: Optional[List[str]] = None,
+        constants_module: Optional[ConstantsModule] = None,
+        zip_includes: Optional[IncludesList] = None,
     ):
-        self.include_files = include_files or []
+        self.include_files: InternalIncludesList = process_path_specs(
+            include_files
+        )
         self.excludes = dict.fromkeys(excludes or [])
         self.optimize_flag = 0
         self.path = path or sys.path
@@ -54,13 +60,13 @@ class ModuleFinder:
         self.zip_exclude_packages = zip_exclude_packages or []
         self.zip_include_packages = zip_include_packages or []
         self.constants_module = constants_module
-        self.zip_includes = zip_includes or []
+        self.zip_includes: InternalIncludesList = process_path_specs(
+            zip_includes
+        )
         self.modules = []
         self.aliases = {}
-        self.exclude_dependent_files = {}
-        self._modules = dict.fromkeys(
-            excludes or []
-        )  # type: Dict[str, Optional[Module]]
+        self.exclude_dependent_files: Set[Path] = set()
+        self._modules: Dict[str, Any] = dict.fromkeys(excludes or [])
         self._builtin_modules = dict.fromkeys(sys.builtin_module_names)
         self._bad_modules = {}
         self._hooks = __import__("cx_Freeze", fromlist=["hooks"]).hooks
@@ -98,7 +104,8 @@ class ModuleFinder:
         """
         Add a module to the list of modules but if one is already found,
         then return it instead; this is done so that packages can be
-        handled properly."""
+        handled properly.
+        """
         module = self._modules.get(name)
         if module is None:
             module = Module(name, path, file_name, parent)
@@ -114,9 +121,9 @@ class ModuleFinder:
                 and module.name not in self.zip_exclude_packages
                 or module.name in self.zip_include_packages
             ):
-                module.store_in_file_system = False
+                module.in_file_system = False
         if module.path is None and path is not None:
-            module.path = path
+            module.path = [Path(p) for p in path]
         if module.file is None and file_name is not None:
             module.file = file_name
         return module
@@ -169,24 +176,19 @@ class ModuleFinder:
         suffixes = importlib.machinery.all_suffixes()
 
         for path in module.path:
-            try:
-                filenames = os.listdir(path)
-            except OSError:
-                continue
-
-            for filename in filenames:
-                fullname = os.path.join(path, filename)
-                if os.path.isdir(fullname):
-                    init_file = os.path.join(fullname, "__init__.py")
-                    if not os.path.exists(init_file):
+            for fullname in path.iterdir():
+                if fullname.is_dir():
+                    init_file = fullname / "__init__.py"
+                    if not init_file.exists():
                         continue
-                    name = filename
+                    name = fullname.name
                 else:
                     # We need to run through these in order to correctly pick
-                    # up PEP 3149 library names (e.g. .cpython-32mu.so).
+                    # up PEP 3149 library names
+                    # (e.g. .cpython-39-x86_64-linux-gnu.so).
                     for suffix in suffixes:
-                        if filename.endswith(suffix):
-                            name = filename[: -len(suffix)]
+                        if fullname.name.endswith(suffix):
+                            name = fullname.name[: -len(suffix)]
 
                             # Skip modules whose names appear to contain '.',
                             # as we may be using the wrong suffix, and even if
@@ -302,7 +304,8 @@ class ModuleFinder:
         """
         Internal method used for importing a module which assumes that the
         name given is an absolute name. None is returned if the module
-        cannot be found."""
+        cannot be found.
+        """
         try:
             # Check in module cache before trying to import it again.
             return self._modules[name]
@@ -350,7 +353,7 @@ class ModuleFinder:
     def _load_module(
         self,
         name: str,
-        path: Union[str, List[str]],
+        path: Union[str, List[str], List[Path]],
         deferred_imports: DeferredList,
         parent: Optional[Module] = None,
     ) -> Optional[Module]:
@@ -363,18 +366,35 @@ class ModuleFinder:
             # Include file as module
             module = self._add_module(name, file_name=path, parent=parent)
             ext = os.path.splitext(os.path.basename(path))[1]
-            if ext in importlib.machinery.SOURCE_SUFFIXES:
+            if ext in importlib.machinery.SOURCE_SUFFIXES + [""]:
                 loader = importlib.machinery.SourceFileLoader(name, path)
             elif ext in importlib.machinery.BYTECODE_SUFFIXES:
                 loader = importlib.machinery.SourcelessFileLoader(name, path)
+            elif ext in importlib.machinery.EXTENSION_SUFFIXES:
+                loader = importlib.machinery.ExtensionFileLoader(name, path)
             else:
-                raise ImportError(f"No module named {name!r}", name=name)
+                loader = None
         else:
             # Find modules to load
+            if path:
+                path = [str(p) for p in path]
             try:
                 # It's recommended to clear the caches first.
                 importlib.machinery.PathFinder.invalidate_caches()
                 spec = importlib.machinery.PathFinder.find_spec(name, path)
+            except KeyError:
+                if parent:
+                    # some packages use a directory with vendored modules
+                    # without an __init__py and are not considered namespace
+                    # packages, then simulate a subpackage
+                    path = [os.path.join(path[0], name.rpartition(".")[-1])]
+                    origin = os.path.join(path[0], "__init__.py")
+                    module = self._add_module(name, path=path, parent=parent)
+                    logging.debug("Adding module [%s] [PACKAGE]", name)
+                    module.code = compile("", origin, "exec")
+                    module.in_import = False
+                    return module
+                spec = None
             except Exception:
                 spec = None
             if spec is None:
@@ -385,7 +405,7 @@ class ModuleFinder:
                 return None
             if loader is importlib.machinery.FrozenImporter:
                 return None
-            # Load package or namespae package
+            # Load package or namespace package
             if spec.submodule_search_locations:
                 path_list = list(spec.submodule_search_locations)
                 module = self._add_module(name, path=path_list, parent=parent)
@@ -393,6 +413,7 @@ class ModuleFinder:
                     logging.debug("Adding module [%s] [NAMESPACE]", name)
                     path = os.path.join(path_list[0], "__init__.py")
                     module.code = compile("", path, "exec")
+                    module.in_import = False
                     return module
                 logging.debug("Adding module [%s] [PACKAGE]", name)
                 path = spec.origin  # path of __init__
@@ -411,7 +432,9 @@ class ModuleFinder:
                 )
             except SyntaxError:
                 logging.debug("Invalid syntax in [%s]", name)
-                raise ImportError(f"Invalid syntax in {path}", name=name)
+                raise ImportError(
+                    f"Invalid syntax in {path}", name=name
+                ) from None
         elif isinstance(loader, importlib.machinery.SourcelessFileLoader):
             logging.debug("Adding module [%s] [BYTECODE]", name)
             # Load Python bytecode
@@ -486,21 +509,26 @@ class ModuleFinder:
         if code is None:
             code = module.code
         # Prepare the new filename.
-        new_filename = original_filename = os.path.normpath(code.co_filename)
+        original_filename = Path(code.co_filename)
         for search_value, replace_value in self.replace_paths:
             if search_value == "*":
                 if top_level_module.file is None:
                     continue
-                search_value = os.path.dirname(top_level_module.file)
                 if top_level_module.path:
-                    search_value = os.path.dirname(search_value)
-                if search_value:
-                    search_value = search_value + os.path.sep
-            if original_filename.startswith(search_value):
-                new_filename = (
-                    replace_value + original_filename[len(search_value) :]
-                )
+                    search_dir = top_level_module.file.parent.parent
+                else:
+                    search_dir = top_level_module.file.parent
+            else:
+                search_dir = Path(search_value)
+            try:
+                new_filename = original_filename.relative_to(search_dir)
+            except ValueError:
+                pass
+            else:
+                new_filename = replace_value / new_filename
                 break
+        else:
+            new_filename = original_filename
 
         # Run on subordinate code objects from function & class definitions.
         consts = list(code.co_consts)
@@ -511,7 +539,7 @@ class ModuleFinder:
                 )
 
         return code_object_replace(
-            code, co_consts=consts, co_filename=new_filename
+            code, co_consts=consts, co_filename=str(new_filename)
         )
 
     def _run_hook(self, hook: str, module_name: str, *args) -> None:
@@ -600,47 +628,66 @@ class ModuleFinder:
         self.aliases[name] = alias_for
 
     def AddConstant(self, name: str, value: str) -> None:
+        """
+        Makes available a constant in the module BUILD_CONSTANTS which is used
+        in the initscripts.
+        """
         self.constants_module.values[name] = value
 
-    def ExcludeDependentFiles(self, filename: str) -> None:
-        self.exclude_dependent_files[filename] = None
+    def ExcludeDependentFiles(self, filename: Union[Path, str]) -> None:
+        self.exclude_dependent_files.add(Path(filename))
 
     def ExcludeModule(self, name: str) -> None:
-        """Exclude the named module from the resulting frozen executable."""
-        self.excludes[name] = None
-        self._modules[name] = None
+        """
+        Exclude the named module and its submodules from the resulting frozen
+        executable."""
+        modules_to_exclude = [name] + [
+            mod for mod in self._modules if mod.startswith(f"{name}.")
+        ]
+        for mod in modules_to_exclude:
+            self.excludes[mod] = None
+            self._modules[mod] = None
 
-    def IncludeFile(self, path: str, name: Optional[str] = None) -> Module:
+    def IncludeFile(
+        self, path: Union[Path, str], name: Optional[str] = None
+    ) -> Module:
         """Include the named file as a module in the frozen executable."""
         if name is None:
-            name = os.path.splitext(os.path.basename(path))[0]
+            name = Path(path).stem
         deferred_imports: DeferredList = []
-        module = self._load_module(name, path, deferred_imports)
+        module = self._load_module(name, str(path), deferred_imports)
         self._import_deferred_imports(deferred_imports)
         return module
 
     def IncludeFiles(
         self,
-        source_path: str,
-        target_path: str,
+        source_path: Union[Path, str],
+        target_path: Union[Path, str],
         copy_dependent_files: bool = True,
     ) -> None:
         """Include the files in the given directory in the target build."""
-        self.include_files.append((source_path, target_path))
+        self.include_files.extend(
+            process_path_specs([(source_path, target_path)])
+        )
         if not copy_dependent_files:
             self.ExcludeDependentFiles(source_path)
 
     def IncludeModule(self, name: str) -> Module:
         """Include the named module in the frozen executable."""
+        # includes has priority over excludes or []
+        if name in self.excludes and name in self._modules:
+            if self._modules[name] is None:
+                self.excludes.pop(name)
+                self._modules.pop(name)
+        # include module
         deferred_imports: DeferredList = []
         module = self._import_module(name, deferred_imports)
         self._import_deferred_imports(deferred_imports, skip_in_import=True)
         return module
 
     def IncludePackage(self, name: str) -> Module:
-        """
-        Include the named package and any submodules in the frozen executable.
-        """
+        """Include the named package and any submodules in the frozen
+        executable."""
         deferred_imports: DeferredList = []
         module = self._import_module(name, deferred_imports)
         if module.path:
@@ -658,12 +705,10 @@ class ModuleFinder:
                 callers = list(self._bad_modules[name].keys())
                 callers.sort()
                 print("? {} imported from {}".format(name, ", ".join(callers)))
-            print(
-                "This is not necessarily a problem - the modules "
-                "may not be needed on this platform.\n"
-            )
+            print("This is not necessarily a problem - the modules ", end="")
+            print("may not be needed on this platform.\n")
 
-    def SetOptimizeFlag(self, optimize_flag):
+    def SetOptimizeFlag(self, optimize_flag: int) -> int:
         """Set a new value of optimize flag and returns the previous value."""
         previous = self.optimize_flag
         # The value of optimize_flag is propagated according to the user's
@@ -673,6 +718,12 @@ class ModuleFinder:
             self.optimize_flag = optimize_flag
         return previous
 
-    def ZipIncludeFiles(self, source_path, target_path):
-        """Include the file(s) in the library.zip"""
-        self.zip_includes.append((source_path, target_path))
+    def ZipIncludeFiles(
+        self,
+        source_path: Union[str, Path],
+        target_path: Optional[Union[str, Path, PurePath]] = None,
+    ) -> None:
+        """Include files or all of the files in a directory to the zip file."""
+        self.zip_includes.extend(
+            process_path_specs([(source_path, target_path)])
+        )
